/*
 ** Module   : FileChecker.java
 ** Abstract : Java implementation of for detecting the file type attribute 
 *             list of OS-DIR
 **
 ** Copyright (c) 2013, Golden Code Development Corporation.
 ** ALL RIGHTS RESERVED. Use is subject to license terms.
 **
 **           Golden Code Development Corporation
 **                      CONFIDENTIAL
 **
 ** -#- -I- --Date-- ----------------Description-----------------
 ** 001 CS  20130103 Created initial version for Implement OS-DIR support as
 *                   an INPUT stream task.
 */
package com.goldencode.p2j.util;

/**
 * <p>
 * Utility class that uses JNI call to get the bitfield representing the 
 * status type of the given file path parameter and return the equivalent
 * Progress OS-DIR status.
 * <p>
 * The following codes are implemented:
 * <p>
 * <b>PRIMARY TYPES</b>
 * <ul>
 *    <li><b>F</b> - Regular file.
 *    <li><b>D</b> - Directory.
 *    <li><b>S</b> - Special Device
 *    <li><b>X</b> - Unknown type
 * </ul>
 * <p>
 * <b>SECUNDARY (ADDITIVE) TYPES</b>
 * <ul>
 *    <li><b>H</b> - Hidden file.
 *    <li><b>L</b> - Symbolic link.
 *    <li><b>P</b> - Pipe File.
 * </ul>
 * <p>
 * One of the main types will always appear, and can also be accompanied by 
 * one or more of the secondary types.
 * <p>
 * In some cases the additive status can only be added to a specific main type
 * Ex: the Progress behaviour for hidden attribute -> only hidden file are 
 * marked as hidden, hidden directories are only marked as directory.
 * <p>
 * Because the file system may be different from an OS to another (some
 * filetypes are Unix specific) results may be yielded differently, 
 * however this is handled by the native code
 *
 */
public class FileChecker
{
   /** Status code for error */
   private static final int ERROR_TYPE =-1;
   
   /** Status bitfield code for unknown */
   private static final int UNKNOWN_TYPE = 0;
   
   /** Status bitfield code for directory */
   private static final int DIRECTORY_TYPE = 1<<0;    // 000001
   
   /** Status bitfield code for file */
   private static final int FILE_TYPE = 1<<1;         // 000010
   
   /** Status bitfield code for symbolic link */
   private static final int SYMBOLIC_LINK_TYPE = 1<<2; //000100
   
   /** Status bitfield code for devices */
   private static final int DEVICE_TYPE = 1<<3;       // 001000
   
   /** Status bitfield code for pipes */
   private static final int PIPE_TYPE = 1<<4;         // 010000
   
   /** Status bitfield code for hidden */
   private static final int HIDDEN_TYPE = 1<<5;       // 100000
   
   // composed types from main + secundary - only the ones that make sense
   
   /** Status bitfield code for directory link */
   private static final int DIRECTORY_LINK_TYPE = DIRECTORY_TYPE | 
                                                  SYMBOLIC_LINK_TYPE;
   
   /** Status bitfield code for directory link */
   private static final int FILE_LINK_TYPE = FILE_TYPE | SYMBOLIC_LINK_TYPE;
   
   /** Status bitfield code for file hidden  link*/
   private static final int FILE_HIDDEN_TYPE = FILE_TYPE | HIDDEN_TYPE;
   
   /** Status bitfield code for directory link */
   private static final int FILE_HIDDEN_LINK_TYPE = FILE_TYPE | HIDDEN_TYPE | 
                                                    SYMBOLIC_LINK_TYPE;
   
   /** Status bitfield code for file pipe */
   private static final int FILE_PIPE_TYPE = FILE_TYPE | PIPE_TYPE;
   
   // load the native library
   static
   {
      System.loadLibrary("p2j");
   }
   
   /**
    * Native utility method that will return a bitfield code representing the
    * status of the file.
    * 
    * @param   filePath
    *          The <code>String</code> representing the path of the file for
    *          which we want to calculate the status code.
    * 
    * @return  The status bitfield code of the file.
    */
   public native static int checkFileType(String filePath);

   /**
    * This method will call the native code handling the file type detection,
    * interpret the bitfield result and return the status in Progress format.
    * 
    * @param  filePath
    *         The <code>String</code> representing the path of the file for
    *         which we want to calculate the status code.
    *         
    * @return The composed status of the file.
    */
   public static String getFileType(String filePath)
   {

      // Iterate over all commandlinearguments and check if the file is a link
      String type = "";
      int val = checkFileType(filePath);
      //TODO - Re-check Progress Unix behavior for file type, when available.
      switch (val)
      {
         case SYMBOLIC_LINK_TYPE :
            type = "L";
            break;
         case FILE_TYPE :
            type = "F";
            break;
         case DIRECTORY_TYPE :
            type ="D";
            break;
         case DEVICE_TYPE :
            type = "S";
            break;
         case PIPE_TYPE :
            type = "P";
            break;
         case DIRECTORY_LINK_TYPE :
            type = "DL";
            break;
         case FILE_LINK_TYPE :
            type = "FL";
            break;
         case FILE_HIDDEN_TYPE :
            type = "FH";
            break;
         case FILE_HIDDEN_LINK_TYPE :
            type = "FHL";
            break;
         case UNKNOWN_TYPE :
            type = "X";
            break;
         default:
            // no need to do anything error occurred and returning empty is 
            // safer 
            break;
      }
      return type;
   }
}
